charge symmetrize(const charge &Q);
bool on_symm_line(const charge &Q);

//-----------------------------------------------------------------------

bool Zmap::contribute(const charge &Q, double value)
{
   bool result = false;
   charge C = symmetrize(Q);
   iterator f = find(C);
   if (f == end())
   {
	if (value >= 1E-140) insert(Zmap::value_type(C, value));
	result = true;
   }
   else (*f).second += value;
   return result;
}

bool Zmap::contribute_nosymm(const charge& Q, double value)
{
   bool result = false;
   charge C = symmetrize(Q);
   if (C == Q)  // contribute only if we are on the not-mirrored side
   {
      iterator f = find(C);
      if (f == end())
      {
         if (value >= 1E-140) insert(Zmap::value_type(C, value));
         result = true;
      }
      else (*f).second += value;
   }
   return result;
}

double Zmap::get_Z(const charge &Q) const
{
   const_iterator f = find(symmetrize(Q));
   return (f == end()) ? 0 : (*f).second;
}

charge symmetrize(const charge &Q)
{
   charge result(Q);
   if (result.qiso < 0) result *= -1;
   return result;
}

bool on_symm_line(const charge &Q)
{
   return (Q.qiso == 0);
}

//-----------------------------------------------------------------


partition_function::partition_function(const particle_zoo &particles,int maxA) :
     its_particle_zoo(particles),
     its_max_A(maxA)
{

  its_charge_boundaries.resize(its_max_A + 1);
  its_Z_map_array.resize(its_max_A + 1);

  particle_zoo::const_iterator i;
  for (i = its_particle_zoo.begin(); i != its_particle_zoo.end(); i++)
    its_Z_map_array[i->its_A].contribute_nosymm(i->its_charge, i->its_W);
  do_computation();
}

partition_function::partition_function(particle_zoo &particles, 
		                       string filename)
   : its_particle_zoo(particles),
     its_inputfilestream(filename.c_str())
{
   if (!its_inputfilestream.good()) throw XFileNotFound();	
   its_max_A = read_headers(particles);
   its_charge_boundaries.resize(its_max_A + 1);
   its_Z_map_array.resize(its_max_A + 1);
   read_data();
}

int partition_function::read_headers(particle_zoo &zoo)
{
  char buf[0xffff];
  int max_A;
  zoo.clear();
  its_inputfilestream.getline(buf, sizeof(buf));	
  assert(strcmp(buf, "PARTITION_FUNCTION CACHE FILE") == 0);
  int t=10;
  while ((t-- > 0) && strcmp(buf, "[env]") != 0)
    its_inputfilestream.getline(buf, sizeof(buf));
   
  assert(strcmp(buf, "[env]") == 0);
  environment env; int particlecounter;
  its_inputfilestream >> env.its_temperature;
  its_inputfilestream >> env.its_volume;
  zoo.set_env_conditions(env);
  its_inputfilestream >> particlecounter;
  its_inputfilestream >> max_A;
  its_inputfilestream.getline(buf, sizeof(buf));

  t = 10; 
  while ((t-- > 0) && strcmp(buf, "[particlezoo]") != 0)
    its_inputfilestream.getline(buf, sizeof(buf));
	   
  assert(strcmp(buf, "[particlezoo]") == 0);
  particle a_particle;
  for (int p = 1; p <= particlecounter; p++)
    {
      its_inputfilestream >> a_particle.its_A;
      its_inputfilestream >> a_particle.its_Mass;
      its_inputfilestream >> a_particle.its_W;
      its_inputfilestream >> a_particle.its_Name;
      its_inputfilestream >> a_particle.its_charge.qiso;
      its_inputfilestream >> a_particle.its_charge.qb;
      its_inputfilestream >> a_particle.its_charge.qs;
      zoo.push_back(a_particle);
    }
  if (its_inputfilestream.eof()) throw XUnexpectedFileError();
  return max_A;
}

void partition_function::read_data()
{
  char buf[0xffff];
  buf[0]=0;
  int t = 10;
  while ((t-- > 0) && (strcmp(buf, "[data]") != 0))
    its_inputfilestream.getline(buf, sizeof(buf));
   
  assert(strcmp(buf, "[data]") == 0);
  while (!its_inputfilestream.eof())
    {
      int A; charge ch; double W;
      its_inputfilestream >> A;
      its_inputfilestream >> ch.qiso;
      its_inputfilestream >> ch.qb;
      its_inputfilestream >> ch.qs;
      its_inputfilestream >> W;
      if (!its_inputfilestream.eof()) 
	its_Z_map_array[A].insert(Zmap::value_type(ch, W));
    }
}

double partition_function::getZ(int A, const charge& Q) const
{
  if ((A > its_max_A)||(A < 0)) throw XOutOfRange();
  if (A == 0)
    {
      charge B(Q);
      B *= -1;
      return ((B == Q) ? 1 : 0);
    }
  else   
    return its_Z_map_array[A].get_Z(Q);   
}

partition_function::~partition_function()
{
}

void partition_function::do_computation()
{
  for (int A = 1; A <= its_max_A; A++)
    {
      cout << "Finished Calculating Z[" << A << "]\n";
#ifdef _DEBUGMSG_	   
      cout << "A = " << A << endl;
#endif
      Zmap::const_iterator i;
      for (i = its_Z_map_array[A].begin(); i != its_Z_map_array[A].end(); i++)
	{
	  particle_zoo::const_iterator particle;
	  for (particle = its_particle_zoo.begin(); 
	       particle != its_particle_zoo.end(); particle++)
	    {
	      charge Q = (*i).first;
	      Q += particle->its_charge;
	      int A_prime = A + particle->its_A;
	      if (A_prime <= its_max_A)
		{
		  double contribution = (*i).second * particle->its_W 
		    * particle->its_A/A_prime; 
		  bool C_onSy = on_symm_line((*i).first);
		  bool Q_onSy = on_symm_line(Q);
		  if (C_onSy && !Q_onSy) 
		    its_Z_map_array[A_prime].contribute_nosymm(Q, contribution);
		  else
		    { 
		      if (Q_onSy && !C_onSy)
			{
			  its_Z_map_array[A_prime].contribute(Q, contribution);
			  Q *= -1;  // give mirror element contribution
			  its_Z_map_array[A_prime].contribute(Q, contribution);
			} else
			  {		
			    its_Z_map_array[A_prime].contribute(Q, contribution);
			  }
		    }  
		}
	    }   // add contributions of all existing particles
	}  // do all existing entries 
    }
}

void partition_function::writeToFile(string filename) const
{
  ofstream outfile(filename.c_str());
  outfile << "PARTITION_FUNCTION CACHE FILE" << endl;
  outfile << "[env]" << endl;
  environment env(its_particle_zoo.get_env_conditions());
  outfile << env.its_temperature << endl;
  outfile << env.its_volume << endl;
  int c = 0;
  particle_zoo::const_iterator i;
  for (i = its_particle_zoo.begin(); i != its_particle_zoo.end(); i++ , c++);
  outfile << c << endl; // write # of particles into env section
  outfile << its_max_A << endl;
  outfile << "[particlezoo]" << endl;
  for (i = its_particle_zoo.begin(); i != its_particle_zoo.end(); i++)
    {
      outfile << i->its_A << "  " << i->its_Mass << "  " << i->its_W << "  "
	      << i->its_Name << endl;
      outfile << "       " << i->its_charge.qiso << " " 
	      << i->its_charge.qb << " " << i->its_charge.qs << endl;
    }	  
  outfile << "[data]" << endl;
  for (int A = 1; A <= its_max_A; A++)
    {
      Zmap::const_iterator j;	  
      for (j=its_Z_map_array[A].begin(); j != its_Z_map_array[A].end(); j++)
	outfile << A << "   "<< j->first.qiso << " " << j->first.qb << " " 
		<< j->first.qs << " " 
		<< setiosflags(ios::scientific) << setprecision(16) 
		<< j->second << endl;
    }
}

void partition_function::calculate_Aprob(charge& Q,double *Aprob){
  double Ztot;
  int a;
  Ztot=0.0;
  for(a=0;a<=its_max_A;a++){
    Aprob[a]=getZ(a,Q);
    Ztot+=Aprob[a];
  }
  cout << "  a       Aprob\n";
  for(a=0;a<=its_max_A;a++){
    Aprob[a]=Aprob[a]/Ztot;
    // cout << setw(3) << a << "  " << setw(10) << Aprob[a] << endl;
  }
}
